;; Generated package description from agitate.el  -*- no-byte-compile: t -*-
(define-package "agitate" "0.0.20241021.65229" "Extras for diff-mode, vc-git, log-edit, log-view" '((emacs "28.1")) :commit "29670e56b6ce227be8249fde906c1d7eeb96d3aa" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "version control" "git") :url "https://github.com/protesilaos/agitate")
